/* --------------------------------------------------------------------------
 *
 * Copyright (C) 2007 Leif Erik Larsen, Kjerringvik, Norway.
 *
 * This file is part of the Open Source Edition of Larsen Commander, as
 * available from http://home.online.no/~leifel/lcmd/.  This code is free 
 * software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License version 3 only, as published by the 
 * Free Software Foundation.  
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 3 at http://www.gnu.org/licenses/gpl-3.0.txt for more details 
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * ------------------------------------------------------------------------ */

#ifndef __GLIB_INTEGER
#define __GLIB_INTEGER

#include "glib/primitives/GIntegerNumber.h"

/**
 * A general integer representation class.
 *
 * @author  Leif Erik Larsen
 * @since   1999.09.05
 */
class GInteger : public GIntegerNumber<int>
{
   private:

      int value;

   public:

      static const int MIN_VALUE;
      static const int MAX_VALUE;

   public:

      GInteger ();
      GInteger ( const GInteger& val );
      GInteger ( int val );
      explicit GInteger ( const char* str, short radix = 10 );

      virtual ~GInteger ();

   public:

      operator int () const { return value; }

   public:

      virtual bool boolValue () const { return value == 0 ? false : true; }
      virtual char charValue () const { return char(value); }
      virtual short shortValue () const { return short(value); }
      virtual int intValue () const { return value; }
      virtual longlong longValue () const { return value; }
      virtual unsigned short ushortValue () const { return (unsigned short) value; }
      virtual unsigned int uintValue () const { return (unsigned int) value; }
      virtual ulonglong ulongValue () const { return ulonglong(value); }
      virtual float floatValue () const { return float(value); }
      virtual double doubleValue () const { return value; }

      virtual int hashCode () const;
      virtual bool equals ( const GObject& obj ) const;

   public:

      static bool IsInteger ( const GString& str, short radix );

      static bool IsInteger ( const GString& str );

      /**
       * Get the parsed integer value of the specified string, assuming the
       * string is specifying an integer with the specified radix.
       *
       * @param   str    String of which to parse.
       * @param   radix  0, 2, 8, 10 or 16. If 0 is specified then we
       *                 will use the default radix as with respect to
       *                 the format of the string; assuming that the
       *                 string is specifying an integer number with
       *                 radix 10, or radix 16 if the string starts
       *                 with the character sequence "0x".
       * @return  The parsed integer value.
       * @throws  GNumberFormatException if the specified string is not
       *                                 possible to parse into an integer.
       */
      static int ParseInt ( const GString& str, short radix = 0 );

      /**
       * Return a <i>GString</i> representation of the current
       * integer value.
       */
      virtual GString toString () const;

      /**
       * Convert the specified integer value to a string.
       *
       * @since  1993.06.01
       * @author Leif Erik Larsen
       * @param  value   Integer value of which to convert.
       * @param  radix   2, 8, 10 or 16.
       * @param  width   Width of the returned string.
       * @param  padChar If the value it self is not large enough to produce
       *                 a string that is at least <i>width</i> characters
       *                 long then we will insert a number of this character
       *                 to satisfy the requested width.
       */
      static GString ToString ( int value, 
                                short radix = 10, 
                                int width = 0, 
                                char padChar = '0' );
};

#endif


